"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generic = exports.parameterProperties = exports.noop = exports.type = exports.ignore = exports.mergeDecorator = exports.decorate = exports.decorateClass = exports.decorateProperty = exports.decorateMethod = exports.decorateParameter = void 0;
require("reflect-metadata");
const helpers_1 = require("./helpers");
const metadata_1 = require("./metadata");
const types_1 = require("./types");
function decorateParameter(data, option) {
    return (target, name, index) => {
        const isCtorParam = helpers_1.metadata.isConstructor(target);
        const targetClass = isCtorParam ? target : target.constructor;
        const methodName = isCtorParam ? "constructor" : name;
        const meta = typeof data === "function" ? data(targetClass, methodName, index) : data;
        metadata_1.setMetadata(Object.assign(Object.assign({}, meta), { [types_1.DecoratorOptionId]: Object.assign(Object.assign({}, meta[types_1.DecoratorOptionId]), option) }), targetClass, methodName, index);
    };
}
exports.decorateParameter = decorateParameter;
function decorateMethod(data, option) {
    return (target, name) => {
        const targetClass = target.constructor;
        const meta = typeof data === "function" ? data(targetClass, name) : data;
        metadata_1.setMetadata(Object.assign(Object.assign({}, meta), { [types_1.DecoratorOptionId]: Object.assign(Object.assign({}, meta[types_1.DecoratorOptionId]), option) }), targetClass, name);
    };
}
exports.decorateMethod = decorateMethod;
function decorateProperty(data, option) {
    return (target, name, index) => {
        if (typeof index === "number")
            decorateParameter(data, option)(target, name, index);
        else
            decorateMethod(data, option)(target, name, index);
    };
}
exports.decorateProperty = decorateProperty;
function decorateClass(data, option) {
    return (target) => {
        const meta = typeof data === "function" ? data(target) : data;
        metadata_1.setMetadata(Object.assign(Object.assign({}, meta), { [types_1.DecoratorOptionId]: Object.assign(Object.assign({}, meta[types_1.DecoratorOptionId]), option) }), target);
    };
}
exports.decorateClass = decorateClass;
function decorate(data, targetTypes = [], option) {
    const throwIfNotOfType = (target) => {
        if (targetTypes.length > 0 && !targetTypes.some(x => x === target))
            throw new Error(`Reflect Error: Decorator of type ${targetTypes.join(", ")} applied into ${target}`);
    };
    return (...args) => {
        //class decorator
        if (args.length === 1) {
            throwIfNotOfType("Class");
            return decorateClass(data, option)(args[0]);
        }
        //parameter decorator
        if (args.length === 3 && typeof args[2] === "number") {
            throwIfNotOfType("Parameter");
            return decorateParameter(data, option)(args[0], args[1], args[2]);
        }
        //property
        if (args[2] === undefined || args[2].get || args[2].set) {
            throwIfNotOfType("Property");
            return decorateProperty(data, option)(args[0], args[1], args[2]);
        }
        throwIfNotOfType("Method");
        return decorateMethod(data, option)(args[0], args[1], args[2]);
    };
}
exports.decorate = decorate;
function mergeDecorator(...fn) {
    return (...args) => {
        fn.forEach(x => x(...args));
    };
}
exports.mergeDecorator = mergeDecorator;
const symIgnore = Symbol("ignore");
const symOverride = Symbol("override");
const symParamProp = Symbol("paramProp");
const symNoop = Symbol("noop");
function ignore() {
    return decorate({ [types_1.DecoratorId]: symIgnore, kind: "Ignore" }, ["Parameter", "Method", "Property"], { allowMultiple: false });
}
exports.ignore = ignore;
function type(type, ...genericParams) {
    return decorate((target) => ({ [types_1.DecoratorId]: symOverride, kind: "Override", type, genericParams: genericParams, target }), ["Parameter", "Method", "Property"], { inherit: true, allowMultiple: false });
}
exports.type = type;
function noop() {
    // type is not inheritable because derived class can define their own type override
    return decorate((target) => ({ [types_1.DecoratorId]: symNoop, kind: "Noop", target }), undefined, { inherit: false, allowMultiple: false });
}
exports.noop = noop;
function parameterProperties() {
    return decorateClass({ [types_1.DecoratorId]: symParamProp, type: "ParameterProperties" }, { allowMultiple: false });
}
exports.parameterProperties = parameterProperties;
var generic;
(function (generic) {
    const symGenericType = Symbol("genericType");
    const symGenericTemplate = Symbol("genericTemplate");
    function template(...templates) {
        return decorateClass(target => ({ [types_1.DecoratorId]: symGenericTemplate, kind: "GenericTemplate", templates, target }), { inherit: false, allowMultiple: false });
    }
    generic.template = template;
    function type(...types) {
        return decorateClass(target => ({ [types_1.DecoratorId]: symGenericType, kind: "GenericType", types, target }), { inherit: false, allowMultiple: false });
    }
    generic.type = type;
    /**
     * Create generic class dynamically
     * @param parent Super class that the class inherited from
     * @param params List of generic type parameters
     */
    function create(parent, ...params) {
        const opt = (typeof parent === "object") ? parent : { parent: parent, name: "DynamicType" };
        const Type = helpers_1.createClass(Object.assign(Object.assign({}, opt), { genericParams: params }));
        return Type;
    }
    generic.create = create;
})(generic = exports.generic || (exports.generic = {}));
